@@ -133,6 +133,7 @@ class TestFileUtils(unittest.TestCase):
         subdir_name = 'aaa'
         subdir = os.path.join(self.tmpdir, subdir_name)
         os.mkdir(subdir)
+        # CWD restored in tearDown
         os.chdir(self.tmpdir)
 
         fname = 'foo.py'
@@ -205,10 +206,10 @@ class TestFileUtils(unittest.TestCase):
             os.path.join(subdir,subdir_name)
         )
 
-    def test_find_library(self):
-        self.tmpdir = os.path.abspath(tempfile.mkdtemp())
-        os.chdir(self.tmpdir)
-
+    @unittest.skipIf(sys.version_info[:2] < (3, 8)
+                     and platform.mac_ver()[0].startswith('10.16'),
+                     "find_library has known bugs in Big Sur for Python<3.8")
+    def test_find_library_system(self):
         # Find a system library (before we muck with the PATH)
         _args = {'cwd':False, 'include_PATH':False, 'pathlist':[]}
         if FileDownloader.get_sysinfo()[0] == 'windows':
@@ -230,6 +231,11 @@ class TestFileUtils(unittest.TestCase):
         _lib = ctypes.cdll.LoadLibrary(a)
         self.assertIsNotNone(_lib)
 
+    def test_find_library_user(self):
+        self.tmpdir = os.path.abspath(tempfile.mkdtemp())
+        # CWD restored in tearDown
+        os.chdir(self.tmpdir)
+
         envvar.PYOMO_CONFIG_DIR = self.tmpdir
         config_libdir = os.path.join(self.tmpdir, 'lib')
         os.mkdir(config_libdir)
@@ -325,6 +331,7 @@ class TestFileUtils(unittest.TestCase):
 
     def test_find_executable(self):
         self.tmpdir = os.path.abspath(tempfile.mkdtemp())
+        # CWD restored in tearDown
         os.chdir(self.tmpdir)
 
         envvar.PYOMO_CONFIG_DIR = self.tmpdir
