@@ -12,11 +12,8 @@ from pyomo.core.base import SymbolMap, NumericLabeler, TextLabeler
 from pyomo.common.timing import HierarchicalTimer
 from pyomo.core.kernel.objective import minimize, maximize
 from .config import WriterConfig
-from .cmodel_converter import PyomoToCModelWalker
 from ..cmodel import cmodel, cmodel_available
 
-id = id
-
 
 class LPWriter(PersistentBase):
     def __init__(self):
@@ -33,7 +30,7 @@ class LPWriter(PersistentBase):
         self._solver_var_to_pyomo_var_map = dict()
         self._solver_con_to_pyomo_con_map = dict()
         self._pyomo_param_to_solver_param_map = dict()
-        self._walker = PyomoToCModelWalker(self._pyomo_var_to_solver_var_map, self._pyomo_param_to_solver_param_map)
+        self._expr_types = None
 
     @property
     def config(self):
@@ -50,6 +47,7 @@ class LPWriter(PersistentBase):
         self.config = saved_config
         self.update_config = saved_update_config
         self._model = model
+        self._expr_types = cmodel.PyomoExprTypes()
 
         if self.config.symbolic_solver_labels:
             self._var_labeler = TextLabeler()
@@ -69,28 +67,10 @@ class LPWriter(PersistentBase):
             self.set_objective(None)
 
     def _add_variables(self, variables: List[_GeneralVarData]):
-        cvars = cmodel.create_vars(len(variables))
-        for ndx, v in enumerate(variables):
-            cv = cvars[ndx]
-            cv.name = self._symbol_map.getSymbol(v, self._var_labeler)
-            if v.is_binary():
-                cv.domain = 'binary'
-            elif v.is_integer():
-                cv.domain = 'integer'
-            else:
-                assert v.is_continuous(), 'LP writer only supports continuous, binary, and integer variables'
-                cv.domain = 'continuous'
-            _, lb, ub, v_is_fixed, v_domain, v_value = self._vars[id(v)]
-            if lb is not None:
-                cv.lb = lb
-            if ub is not None:
-                cv.ub = ub
-            if v_value is not None:
-                cv.value = v_value
-            if v_is_fixed:
-                cv.fixed = True
-            self._pyomo_var_to_solver_var_map[id(v)] = cv
-            self._solver_var_to_pyomo_var_map[cv] = v
+        cmodel.process_pyomo_vars(self._expr_types, variables, self._pyomo_var_to_solver_var_map,
+                                  self._pyomo_param_to_solver_param_map, self._vars,
+                                  self._solver_var_to_pyomo_var_map, True, self._symbol_map,
+                                  self._var_labeler, False)
 
     def _add_params(self, params: List[_ParamData]):
         cparams = cmodel.create_params(len(params))
@@ -133,31 +113,10 @@ class LPWriter(PersistentBase):
             self._param_labeler.remove_obj(p)
 
     def _update_variables(self, variables: List[_GeneralVarData]):
-        for v in variables:
-            cv = self._pyomo_var_to_solver_var_map[id(v)]
-            if v.is_binary():
-                cv.domain = 'binary'
-            elif v.is_integer():
-                cv.domain = 'integer'
-            else:
-                assert v.is_continuous(), 'LP writer only supports continuous, binary, and integer variables'
-                cv.domain = 'continuous'
-            lb = value(v.lb)
-            ub = value(v.ub)
-            if lb is None:
-                cv.lb = -cmodel.inf
-            else:
-                cv.lb = lb
-            if ub is None:
-                cv.ub = cmodel.inf
-            else:
-                cv.ub = ub
-            if v.value is not None:
-                cv.value = v.value
-            if v.is_fixed():
-                cv.fixed = True
-            else:
-                cv.fixed = False
+        cmodel.process_pyomo_vars(self._expr_types, variables, self._pyomo_var_to_solver_var_map,
+                                  self._pyomo_param_to_solver_param_map, self._vars,
+                                  self._solver_var_to_pyomo_var_map, False, None,
+                                  None, True)
 
     def update_params(self):
         for p_id, p in self._params.items():
@@ -165,32 +124,18 @@ class LPWriter(PersistentBase):
             cp.value = p.value
 
     def _set_objective(self, obj: _GeneralObjectiveData):
+        cobj = cmodel.process_lp_objective(self._expr_types, obj, self._pyomo_var_to_solver_var_map,
+                                           self._pyomo_param_to_solver_param_map)
         if obj is None:
-            const = cmodel.Constant(0)
-            lin_coef = list()
-            lin_vars = list()
-            quad_coef = list()
-            quad_vars_1 = list()
-            quad_vars_2 = list()
             sense = 0
+            cname = 'objective'
         else:
-            repn = generate_standard_repn(obj.expr, compute_values=False, quadratic=True)
-            const = self._walker.dfs_postorder_stack(repn.constant)
-            lin_coef = [self._walker.dfs_postorder_stack(i) for i in repn.linear_coefs]
-            lin_vars = [self._pyomo_var_to_solver_var_map[id(i)] for i in repn.linear_vars]
-            quad_coef = [self._walker.dfs_postorder_stack(i) for i in repn.quadratic_coefs]
-            quad_vars_1 = [self._pyomo_var_to_solver_var_map[id(i[0])] for i in repn.quadratic_vars]
-            quad_vars_2 = [self._pyomo_var_to_solver_var_map[id(i[1])] for i in repn.quadratic_vars]
+            cname = self._symbol_map.getSymbol(obj, self._obj_labeler)
             if obj.sense is minimize:
                 sense = 0
             else:
                 sense = 1
-        cobj = cmodel.LPObjective(const, lin_coef, lin_vars, quad_coef, quad_vars_1, quad_vars_2)
         cobj.sense = sense
-        if obj is None:
-            cname = 'objective'
-        else:
-            cname = self._symbol_map.getSymbol(obj, self._obj_labeler)
         cobj.name = cname
         self._writer.objective = cobj
 
