@@ -485,16 +485,35 @@ E : Size=2
         self.assertEqual(len(model.E), 2)
 
     def test_implicit_definition(self):
+        model = ConcreteModel()
+        model.idx = Set(initialize=[1,2,3])
+        model.E = Expression(model.idx)
+        self.assertEqual(len(model.E), 3)
+        expr = model.E[1]
+        self.assertIs(type(expr), _GeneralExpressionData)
+        model.E[1] = None
+        self.assertIs(expr, model.E[1])
+        self.assertIs(type(expr), _GeneralExpressionData)
+        self.assertIs(expr.expr, None)
+        model.E[1] = 5
+        self.assertIs(expr, model.E[1])
+        self.assertEqual(model.E.extract_values(), {1:5, 2:None, 3:None})
+        model.E[2] = 6
+        self.assertIsNot(expr, model.E[2])
+        self.assertEqual(model.E.extract_values(), {1:5, 2:6, 3:None})
+
+    def test_explicit_skip_definition(self):
         model = ConcreteModel()
         model.idx = Set(initialize=[1,2,3])
         model.E = Expression(model.idx, rule=lambda m,i: Expression.Skip)
         self.assertEqual(len(model.E), 0)
-        expr = model.E[1]
-        self.assertIsNone(expr)
+        with self.assertRaises(KeyError):
+            expr = model.E[1]
+
         model.E[1] = None
         expr = model.E[1]
         self.assertIs(type(expr), _GeneralExpressionData)
-        self.assertIs(expr.value, None)
+        self.assertIs(expr.expr, None)
         model.E[1] = 5
         self.assertIs(expr, model.E[1])
         self.assertEqual(model.E.extract_values(), {1:5})
