@@ -18,7 +18,7 @@ from weakref import ref as weakref_ref
 from pyomo.common.deprecation import RenamedClass,  deprecation_warning
 from pyomo.common.errors import PyomoException
 from pyomo.common.log import is_debug_set
-from pyomo.common.modeling import unique_component_name
+from pyomo.common.modeling import unique_component_name, NOTSET
 from pyomo.common.timing import ConstructionTimer
 from pyomo.core import (
     ModelComponentFactory, Binary, Block, ConstraintList, Any,
@@ -68,23 +68,23 @@ class AutoLinkedBinaryVar(ScalarVar):
     def get_associated_boolean(self):
         return self._associated_boolean()
 
-    def set_value(self, val, valid=False):
-        super().set_value(val, valid)
-        bool_var = self.get_associated_boolean()
-        # Only update the associated Boolean value if it is needed
-        # to match the current (potentially fractional) binary value.
-        # (This prevents infinite recursion.)
+    def set_value(self, val, skip_validation=False, _propagate_value=True):
+        super().set_value(val, skip_validation)
+        if not _propagate_value:
+            return
+        # Map the incoming (numeric) value to bool/None
         if val is None:
             bool_val = None
         elif fabs(val - 0.5) < 0.5 - AutoLinkedBinaryVar.INTEGER_TOLERANCE:
             bool_val = None
         else:
             bool_val = bool(int(val + 0.5))
-        if bool_val != bool_var.value:
-            bool_var.set_value(bool_val)
+        # (Setting _propagate_value prevents infinite recursion.)
+        self.get_associated_boolean().set_value(
+            bool_val, skip_validation, _propagate_value=False)
 
-    def fix(self, *val):
-        super().fix(*val)
+    def fix(self, value=NOTSET, skip_validation=False):
+        super().fix(value, skip_validation)
         bool_var = self.get_associated_boolean()
         if not bool_var.is_fixed():
             bool_var.fix()
@@ -147,45 +147,26 @@ class AutoLinkedBooleanVar(ScalarBooleanVar):
             % (self.name,), version='6.0')
         return self.get_associated_binary()
 
-    @property
-    def value(self):
-        return self._value
-
-    @value.setter
-    def value(self, val):
+    def set_value(self, val, skip_validation=False, _propagate_value=True):
         # super() does not work as expected for properties; we will call
         # the property setter explicitly.
-        ScalarBooleanVar.value.fset(self, val)
-        bin_var = self.get_associated_binary()
-        bin_val = bin_var.value
-        if bin_val is None:
-            bool_val = None
-        elif fabs(bin_val - 0.5) < 0.5 - AutoLinkedBinaryVar.INTEGER_TOLERANCE:
-            bool_val = None
-        else:
-            bool_val = bool(int(bin_val + 0.5))
-        # Fetch the current value (so that it is cast to None/bool)
-        val = self.value
-        # Only update the associated (potentially fractional) binary
-        # value if it is needed to match the current Boolean value.
-        # (This prevents infinite recursion.)
-        if val != bool_val:
-            if val is not None:
-                val = int(val)
-            bin_var.set_value(val)
-
-    def fix(self, *val):
-        super().fix(*val)
-        bin_var = self.get_associated_binary()
-
+        super().set_value(val, skip_validation)
+        if not _propagate_value:
+            return
+        # Fetch the current value (so we know it has already been cast
+        # to None/bool)
         val = self.value
         if val is not None:
             val = int(val)
-        if not bin_var.is_fixed() or bin_var.value != val:
-            # Note: if someone fixes the Boolean to True/False then we
-            # need to snap the binary to 1/0 (and not leave it at the
-            # potentially fractional value)
-            bin_var.fix(val)
+        # (Setting _propagate_value prevents infinite recursion.)
+        self.get_associated_binary().set_value(
+            val, skip_validation, _propagate_value=False)
+
+    def fix(self, value=NOTSET, skip_validation=False):
+        super().fix(value, skip_validation)
+        bin_var = self.get_associated_binary()
+        if not bin_var.is_fixed():
+            bin_var.fix()
 
     def unfix(self):
         super().unfix()
