@@ -21,6 +21,7 @@ from pyomo.common.tee import capture_output
 from pyomo.common.tempfiles import TempfileManager
 from pyomo.core.expr.numvalue import is_fixed
 from pyomo.core.expr.numvalue import value
+from pyomo.core.staleflag import StaleFlagManager
 from pyomo.repn import generate_standard_repn
 from pyomo.solvers.plugins.solvers.direct_solver import DirectSolver
 from pyomo.solvers.plugins.solvers.direct_or_persistent_solver import DirectOrPersistentSolver
@@ -163,14 +164,7 @@ class XpressDirect(DirectSolver):
         return bool(xpress_available)
 
     def _apply_solver(self):
-        if not self._save_results:
-            for block in self._pyomo_model.block_data_objects(descend_into=True,
-                                                              active=True):
-                for var in block.component_data_objects(ctype=pyomo.core.base.var.Var,
-                                                        descend_into=False,
-                                                        active=True,
-                                                        sort=False):
-                    var.stale = True
+        StaleFlagManager.mark_all_as_stale()
 
         self._solver_model.setlogfile(self._log_file)
         if self._keepfiles:
@@ -689,7 +683,6 @@ class XpressDirect(DirectSolver):
                 for xpress_var, val in zip(xpress_vars, var_vals):
                     pyomo_var = self._solver_var_to_pyomo_var_map[xpress_var]
                     if self._referenced_variables[pyomo_var] > 0:
-                        pyomo_var.stale = False
                         soln_variables[xpress_var.name] = {"Value": val}
 
                 if extract_reduced_costs:
@@ -731,7 +724,7 @@ class XpressDirect(DirectSolver):
             if xprob_attrs.lpstatus == xp.lp_optimal and \
                     ((not is_mip) or (xprob_attrs.mipsols > 0)):
 
-                self._load_vars()
+                self.load_vars()
 
                 if extract_reduced_costs:
                     self._load_rc()
@@ -772,8 +765,7 @@ class XpressDirect(DirectSolver):
 
         for var, val in zip(vars_to_load, vals):
             if ref_vars[var] > 0:
-                var.stale = False
-                var.value = val
+                var.set_value(val, skip_validation=True)
 
     def _load_rc(self, vars_to_load=None):
         if not hasattr(self._pyomo_model, 'rc'):
